import importlib
from dataclasses import dataclass
from typing import Final, Optional, Type

from scrapy.crawler import Crawler

from scrapy_llm.types import T

LLM_EXTRACTED_DATA_KEY: Final[str] = "llm_extracted_data"
LLM_RESPONSE_MODEL_KEY: Final[str] = "llm_response_model"


@dataclass
class LlmExtractorConfig:
    response_model: Type[T]
    unwrap_nested: bool = False
    llm_api_base: str = "https://api.openai.com/v1"
    llm_model: str = "gpt-4-turbo"
    llm_temperature: float = 0.0001
    llm_system_message: str = (
        "You are a data extraction expert, your role is to extract data from the given text according to the provided schema. Make sure your output is a valid JSON object. "
        "You are crawling the following URL: {url}"
    )
    llm_additional_system_message: str = ""
    html_cleaner_ignore_links: bool = True
    html_cleaner_ignore_images: bool = True

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> "LlmExtractorConfig":
        api_base: str = crawler.settings.get("LLM_API_BASE", cls.llm_api_base)
        system_message: str = crawler.settings.get(
            "LLM_SYSTEM_MESSAGE", cls.llm_system_message
        )
        additional_system_message: str = crawler.settings.get(
            "LLM_ADDITIONAL_SYSTEM_MESSAGE", ""
        )
        model: str = crawler.settings.get("LLM_MODEL", cls.llm_model)
        model_temperature: float = crawler.settings.get(
            "LLM_MODEL_TEMPERATURE", cls.llm_temperature
        )
        response_model_path: str = crawler.settings.get("LLM_RESPONSE_MODEL", None)
        unwrap_nested: bool = crawler.settings.get(
            "LLM_UNWRAP_NESTED", cls.unwrap_nested
        )
        response_model: Optional[Type[T]] = getattr(
            crawler.spider, "response_model", None
        )

        if not response_model and response_model_path:
            module_path, class_name = response_model_path.rsplit(".", 1)
            module = importlib.import_module(module_path)
            response_model = getattr(module, class_name)

        html_cleaner_ignore_links: bool = crawler.settings.get(
            "HTML_CLEANER_IGNORE_LINKS", True
        )
        html_cleaner_ignore_images: bool = crawler.settings.get(
            "HTML_CLEANER_IGNORE_IMAGES", True
        )

        return cls(
            response_model=response_model,
            llm_api_base=api_base,
            unwrap_nested=unwrap_nested,
            llm_model=model,
            llm_temperature=model_temperature,
            llm_system_message=system_message,
            llm_additional_system_message=additional_system_message,
            html_cleaner_ignore_links=html_cleaner_ignore_links,
            html_cleaner_ignore_images=html_cleaner_ignore_images,
        )



from typing import Any, Dict, Iterable

import pandas as pd
import scrapy
from dorm_capacity.utils.models import LLMDormItem
from scrapy import Request
from scrapy.http.response.html import HtmlResponse

from scrapy_llm.config import LLM_EXTRACTED_DATA_KEY


class GenericSpider(scrapy.Spider):
    """
    Generic spider for scraping generic dorm capacity data.

    Args:
        urls_file: csv file containing URLs to scrape.

    expected csv file format:

    ```csv\n
    url
    https://example.com
    https://example2.com
    ```
    """

    name = "generic"
    response_model = LLMDormItem

    def __init__(self, urls_file: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # urls_file is a csv file with a column named "url"
        url_df = pd.read_csv(urls_file, index_col=False)
        self.start_urls = url_df["url"].tolist()

    def start_requests(self) -> Iterable[Request]:
        """
        Start requests for each of `self.start_urls`.
        """
        for url in self.start_urls:
            yield Request(url=url)

    def parse(self, response: HtmlResponse) -> Dict[str, Any]:
        return response.request.meta.get(LLM_EXTRACTED_DATA_KEY)



from typing import Generic, Iterable, List, Type, TypeVar

import instructor
from litellm import completion
from scrapy import Request, signals
from scrapy.crawler import Crawler
from scrapy.http.response import Response
from scrapy.spiders import Spider
from scrapy.utils.log import logger
from scrapy.exceptions import NotConfigured

from scrapy_llm.config import LLM_EXTRACTED_DATA_KEY, LLM_RESPONSE_MODEL_KEY, LlmExtractorConfig
from scrapy_llm.types import LLMOutput, T
from scrapy_llm.utils import flatten_dict, process_html

LLMExtractor = TypeVar("LLMExtractor", bound="LlmExtractorMiddleware")


class LlmExtractorMiddleware(Generic[T]):
    def __init__(
        self,
        crawler: Crawler,
        config: LlmExtractorConfig,
    ):
        self.crawler = crawler
        self.config = config

    @classmethod
    def from_crawler(cls: Type[LLMExtractor], crawler: Crawler) -> LLMExtractor:
        config = LlmExtractorConfig.from_crawler(crawler)
        instance = cls(crawler, config)
        crawler.signals.connect(instance.spider_opened, signal=signals.spider_opened)
        return instance

    def process_response(
        self, request: Request, response: Response, spider: Spider
    ) -> Response:
        response_model = request.meta.get(LLM_RESPONSE_MODEL_KEY) or self.config.response_model
        if not response_model:
            raise NotConfigured(
                """
                Response model not provided for LlmExtractorMiddleware.
                Please set LLM_RESPONSE_MODEL to class path in settings, define response_model in the spider, or provide response model object in Request meta.
                """
            )

        extracted_data = self.extract_item_data(
            response.text, response_model, request.url
        )

        if self.config.unwrap_nested:
            extracted_data = flatten_dict(extracted_data)

        request.meta[LLM_EXTRACTED_DATA_KEY] = extracted_data

        return response

    def spider_opened(self, spider: Spider) -> None:
        spider.logger.info("Spider opened: %s" % spider.name)

    def extract_item_data(
        self,
        raw_html: str,
        response_model: Type[T],
        url: str,
    ) -> List[LLMOutput]:
        """Extract data from the given HTML text using an LLM model."""

        cl = instructor.from_litellm(completion)

        system_message = f"{self.config.llm_system_message.format(url=url)} {self.config.llm_additional_system_message}"

        resp: List[response_model] = cl.chat.completions.create(
            model=self.config.llm_model,
            api_base=self.config.llm_api_base,
            custom_llm_provider="openai",
            temperature=self.config.llm_temperature,
            messages=[
                {
                    "role": "system",
                    "content": system_message,
                },
                {
                    "role": "user",
                    "content": process_html(
                        raw_html,
                        self.config.html_cleaner_ignore_links,
                        self.config.html_cleaner_ignore_images,
                    ),
                },
            ],
            # always return an iterable of the given type to simplify nested type cascading.
            response_model=Iterable[response_model],
        )

        items = [item.model_dump() for item in resp]

        return items
